from collections.abc import Hashable, Mapping, Sequence
from typing import Any, Literal, Optional, Union, cast, overload

from beartype import beartype
from fastparquet import ParquetFile, write
from pandas import DataFrame, Series

from utilities.atomicwrites import writer
from utilities.iterables import is_iterable_not_str
from utilities.numpy import datetime64ns, has_dtype
from utilities.pandas import Int64, category, check_range_index, string
from utilities.pathlib import PathLike

_Compression = Literal["gzip", ",snappy", "brotli", "lz4", "zstandard"]
Compression = Union[_Compression, Mapping[Hashable, Optional[_Compression]]]
_Op = Literal["==", "=", ">", ">=", "<", "<=", "!=", "in", "not in"]
_Filter = tuple[Hashable, _Op, Any]
Filters = Union[Sequence[_Filter], Sequence[Sequence[_Filter]]]
_PARQUET_DTYPES = [bool, datetime64ns, float, Int64, string]


@beartype
def count_rows(path: PathLike, /, *, filters: Optional[Filters] = None) -> int:
    """Count the number of rows in a Parquet file."""
    return _get_parquet_file(path).count(
        filters=filters,
        **_maybe_row_filter(filters),
    )


@beartype
def get_columns(path: PathLike, /) -> list[Hashable]:
    """Get the columns in a Parquet file."""
    return _get_parquet_file(path).columns


@beartype
def get_dtypes(path: PathLike, /) -> dict[Hashable, Any]:
    """Get the dtypes in a Parquet file.

    Note that we store strings as categoricals, so we will report `string`
    instead of category here.
    """
    dtypes = cast(dict[Hashable, Any], _get_parquet_file(path).dtypes)
    return {k: string if v == category else v for k, v in dtypes.items()}


@overload
def read_parquet(
    path: PathLike,
    /,
    *,
    head: Optional[int] = None,
    columns: Hashable,
    filters: Optional[Filters] = None,
) -> Series:
    ...


@overload
def read_parquet(
    path: PathLike,
    /,
    *,
    head: Optional[int] = None,
    columns: Optional[Sequence[Hashable]] = None,
    filters: Optional[Filters] = None,
) -> DataFrame:
    ...


@beartype
def read_parquet(
    path: PathLike,
    /,
    *,
    head: Optional[int] = None,
    columns: Optional[Union[Hashable, Sequence[Hashable]]] = None,
    filters: Optional[Filters] = None,
) -> Union[Series, DataFrame]:
    """Read a Parquet file into a Series/DataFrame."""
    file = _get_parquet_file(path)
    as_df = (columns is None) or is_iterable_not_str(columns)
    columns_use = columns if as_df else [columns]
    kwargs = _maybe_row_filter(filters)
    if head is None:
        df = file.to_pandas(columns=columns_use, filters=filters, **kwargs)
    else:
        df = file.head(head, columns=columns_use, filters=filters, **kwargs)
    dtypes = {
        k: string for k, v in df.items() if has_dtype(v, [category, object])
    }
    df = df.astype(dtypes).reset_index(drop=True)
    return df if as_df else df[columns]


@beartype
def _get_parquet_file(path: PathLike, /) -> ParquetFile:
    """Read a Parquet file."""
    return ParquetFile(path, verify=True)


@beartype
def _maybe_row_filter(filters: Optional[Filters], /) -> dict[str, bool]:
    """Add the `row_filter` argument if necessary."""
    return {} if filters is None else {"row_filter": True}


@beartype
def write_parquet(
    df: DataFrame,
    path: PathLike,
    /,
    *,
    overwrite: bool = False,
    compression: Optional[Compression] = "gzip",
) -> None:
    """Atomically write a DataFrame to a Parquet file."""
    if len(df) == 0:
        msg = f"DataFrame is empty: {df=}"
        raise EmptyDataFrameError(msg)
    check_range_index(df)
    for _, column in df.items():
        if not has_dtype(column, _PARQUET_DTYPES):
            msg = f"Invalid dtype: {column=}"
            raise TypeError(msg)
    dtypes = {k: category for k, v in df.dtypes.items() if v == string}
    df = df.astype(dtypes)
    with writer(path, overwrite=overwrite) as temp:
        write(temp, df, compression=compression)


class EmptyDataFrameError(ValueError):
    """Raised when a DataFrame is empty."""



from beartype.vale import IsAttr, IsEqual

from utilities.numpy import datetime64D, datetime64ns, datetime64Y

# dtype
DTypeB = IsAttr["dtype", IsEqual[bool]]
DTypeDD = IsAttr["dtype", IsEqual[datetime64D]]
DTypeDY = IsAttr["dtype", IsEqual[datetime64Y]]
DTypeDns = IsAttr["dtype", IsEqual[datetime64ns]]
DTypeF = IsAttr["dtype", IsEqual[float]]
DTypeI = IsAttr["dtype", IsEqual[int]]
DTypeO = IsAttr["dtype", IsEqual[object]]

# ndim
NDim0 = IsAttr["ndim", IsEqual[0]]
NDim1 = IsAttr["ndim", IsEqual[1]]
NDim2 = IsAttr["ndim", IsEqual[2]]
NDim3 = IsAttr["ndim", IsEqual[3]]



import datetime as dt
from collections.abc import Iterable, Iterator
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

from beartype import beartype
from luigi import Parameter, Target, Task
from luigi import build as _build
from luigi.interface import LuigiRunResult
from luigi.notifications import smtp
from luigi.parameter import MissingParameterException
from luigi.task import Register, flatten

from utilities.datetime import parse_time, serialize_time
from utilities.enum import ensure_enum, parse_enum
from utilities.logging import LogLevel
from utilities.pathlib import PathLike

_E = TypeVar("_E", bound=Enum)


class EnumParameter(Parameter, Generic[_E]):
    """A parameter which takes the value of an Enum."""

    @beartype
    def __init__(
        self,
        enum: type[_E],
        /,
        *args: Any,
        case_sensitive: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._enum = enum
        self._case_sensitive = case_sensitive

    @beartype
    def normalize(self, member: Union[_E, str], /) -> _E:  # noqa: D102
        return ensure_enum(
            self._enum,
            member,
            case_sensitive=self._case_sensitive,
        )

    @beartype
    def parse(self, member: str, /) -> _E:  # noqa: D102
        return parse_enum(
            self._enum,
            member,
            case_sensitive=self._case_sensitive,
        )

    @beartype
    def serialize(self, member: _E, /) -> str:  # noqa: D102
        return member.name


class TimeParameter(Parameter, Generic[_E]):
    """A parameter which takes the value of a `dt.time`."""

    @beartype
    def normalize(self, time: dt.time, /) -> dt.time:  # noqa: D102
        return time

    @beartype
    def parse(self, time: str, /) -> dt.time:  # noqa: D102
        return parse_time(time)

    @beartype
    def serialize(self, time: dt.time, /) -> str:  # noqa: D102
        return serialize_time(time)


class PathTarget(Target):
    """A local target whose `path` attribute is a Pathlib instance."""

    @beartype
    def __init__(self, path: PathLike, /) -> None:
        super().__init__()
        self.path = Path(path)

    @beartype
    def exists(self) -> bool:
        """Check if the target exists."""
        return self.path.exists()


@overload
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: Literal[False] = False,
    local_scheduler: bool = False,
    log_level: Optional[LogLevel] = None,
    workers: Optional[int] = None,
) -> bool:
    ...


@overload
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: Literal[True],
    local_scheduler: bool = False,
    log_level: Optional[LogLevel] = None,
    workers: Optional[int] = None,
) -> LuigiRunResult:
    ...


@beartype
def build(
    task: Iterable[Task],
    /,
    *,
    detailed_summary: bool = False,
    local_scheduler: bool = False,
    log_level: Optional[LogLevel] = None,
    workers: Optional[int] = None,
) -> Union[bool, LuigiRunResult]:
    """Build a set of tasks."""
    return _build(
        task,
        detailed_summary=detailed_summary,
        local_scheduler=local_scheduler,
        **({} if log_level is None else {"log_level": log_level}),
        **({} if workers is None else {"workers": workers}),
    )


_Task = TypeVar("_Task", bound=Task)


@beartype
def clone(task: Task, cls: type[_Task], /, **kwargs: Any) -> _Task:
    """Clone a task."""
    return cast(_Task, task.clone(cls, **kwargs))


@overload
def get_dependencies_downstream(
    task: Task,
    /,
    *,
    cls: type[_Task],
    recursive: bool = False,
) -> frozenset[_Task]:
    ...


@overload
def get_dependencies_downstream(
    task: Task,
    /,
    *,
    cls: None = None,
    recursive: bool = False,
) -> frozenset[Task]:
    ...


@beartype
def get_dependencies_downstream(
    task: Task,
    /,
    *,
    cls: Optional[type[Task]] = None,
    recursive: bool = False,
) -> frozenset[Task]:
    """Get the downstream dependencies of a task."""
    return frozenset(
        _yield_dependencies_downstream(task, cls=cls, recursive=recursive),
    )


@beartype
def _yield_dependencies_downstream(
    task: Task,
    /,
    *,
    cls: Optional[type[Task]] = None,
    recursive: bool = False,
) -> Iterator[Task]:
    for task_cls in cast(Iterable[type[Task]], get_task_classes(cls=cls)):
        try:
            cloned = clone(task, task_cls)
        except (MissingParameterException, TypeError):
            pass
        else:
            if task in get_dependencies_upstream(cloned, recursive=recursive):
                yield cloned
                if recursive:
                    yield from get_dependencies_downstream(
                        cloned,
                        recursive=recursive,
                    )


@beartype
def get_dependencies_upstream(
    task: Task,
    /,
    *,
    recursive: bool = False,
) -> frozenset[Task]:
    """Get the upstream dependencies of a task."""
    return frozenset(_yield_dependencies_upstream(task, recursive=recursive))


@beartype
def _yield_dependencies_upstream(
    task: Task,
    /,
    *,
    recursive: bool = False,
) -> Iterator[Task]:
    for t in cast(Iterable[Task], flatten(task.requires())):
        yield t
        if recursive:
            yield from get_dependencies_upstream(t, recursive=recursive)


@overload
def get_task_classes(*, cls: type[_Task]) -> frozenset[type[_Task]]:
    ...


@overload
def get_task_classes(*, cls: None = None) -> frozenset[type[Task]]:
    ...


@beartype
def get_task_classes(
    *,
    cls: Optional[type[_Task]] = None,
) -> frozenset[type[_Task]]:
    """Yield the task classes. Optionally filter down."""
    return frozenset(_yield_task_classes(cls=cls))


@beartype
def _yield_task_classes(
    *,
    cls: Optional[type[_Task]] = None,
) -> Iterator[type[_Task]]:
    """Yield the task classes. Optionally filter down."""
    for name in Register.task_names():
        task_cls = Register.get_task_cls(name)
        if (
            (cls is None)
            or ((cls is not task_cls) and issubclass(task_cls, cls))
        ) and (task_cls is not smtp):
            yield cast(type[_Task], task_cls)



from dataclasses import dataclass
from pathlib import Path

from beartype import beartype
from typed_settings import option, settings


@settings(frozen=True)
class Config:
    """Settings for the `rotate_logs` script."""

    path: Path = option(
        default=Path.cwd(),
        click={"param_decls": ("-p", "--path")},
    )
    extension: str = option(
        default="log",
        click={"param_decls": ("-e", "--extension")},
    )
    size: int = option(
        default=int(100 * 1024),
        click={"param_decls": ("-s", "--size")},
    )
    keep: int = option(default=3, click={"param_decls": ("-k", "--keep")})
    dry_run: bool = option(
        default=False,
        click={"param_decls": ("-dr", "--dry-run")},
    )


@beartype
@dataclass(frozen=True)
class Item:
    """A log file, identified."""

    path: Path
    head: Path
    num: int = 0



import datetime as dt
from collections.abc import Iterable, Iterator
from decimal import Decimal
from typing import Any, Optional, Union, overload

from beartype import beartype
from pandas import DataFrame, Series
from sqlalchemy import Column, insert
from sqlalchemy.engine import Connection, Engine, Row
from sqlalchemy.exc import DuplicateColumnError
from sqlalchemy.sql import ColumnElement, Select

from utilities.inflection import snake_case
from utilities.iterables import DuplicatedError, check_duplicates
from utilities.more_itertools import EmptyIterableError, one
from utilities.numpy import datetime64ns, has_dtype
from utilities.pandas import (
    Int64,
    boolean,
    datetime64nsutc,
    string,
    timestamp_to_date,
    timestamp_to_datetime,
)
from utilities.sqlalchemy import (
    get_column_names,
    get_columns,
    get_dialect,
    get_table,
    model_to_dict,
    yield_connection,
)
from utilities.text import ensure_str


@beartype
def insert_dataframe(
    df: DataFrame,
    table_or_model: Any,
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    snake: bool = False,
) -> None:
    """Insert a DataFrame into a database."""
    return insert_items([(df, table_or_model)], engine_or_conn, snake=snake)


@beartype
def insert_items(
    items: Iterable[Any],
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    snake: bool = False,
) -> None:
    """Insert a set of items into a database.

    These can be either a:
     - ([tuple], table) pair, or
     - (DataFrame, table) pair, or
     - Model instance.
    """
    dialect = get_dialect(engine_or_conn)
    with yield_connection(engine_or_conn) as conn:
        for item in items:
            if isinstance(item, tuple):
                first, table_or_model = item
                if isinstance(first, list):
                    values = first
                elif isinstance(first, DataFrame):
                    values = list(
                        _yield_dataframe_rows_as_dicts(
                            first,
                            table_or_model,
                            snake=snake,
                        ),
                    )
                else:
                    msg = f"Invalid type: {first=}"
                    raise TypeError(msg)
            else:
                table_or_model = item
                values = [model_to_dict(item)]
            ins = insert(get_table(table_or_model))
            if len(values) >= 1:
                if dialect == "oracle":  # pragma: no cover
                    _ = conn.execute(ins, values)
                else:
                    _ = conn.execute(ins.values(values))


@beartype
def _yield_dataframe_rows_as_dicts(
    df: DataFrame,
    table_or_model: Any,
    /,
    *,
    snake: bool = False,
) -> Iterator[dict[str, Any]]:
    """Yield the rows of a DataFrame as dicts, ready for insertion."""
    parsed = [
        _parse_series_against_table(sr, table_or_model, snake=snake)
        for _, sr in df.items()
    ]
    keys = [key for key, _ in parsed]
    for row in zip(*(it for _, it in parsed)):
        yield dict(zip(keys, row))


@beartype
def _parse_series_against_table(
    series: Series,
    table_or_model: Any,
    /,
    *,
    snake: bool = False,
) -> tuple[str, Iterator[Any]]:
    """Parse a series against a table.

    In particular:
     - check the column which it will insert into, and
     - yield the elements for insertion.
    """
    series_name = ensure_str(series.name)
    if snake:
        column_names = map(snake_case, get_column_names(table_or_model))
        target_name = snake_case(series_name)
        error = SeriesNameSnakeCaseNotInTableError
    else:
        column_names = get_column_names(table_or_model)
        target_name = snake_case(series_name)
        error = SeriesNameNotInTableError
    try:
        column = one(
            col
            for name, col in zip(column_names, get_columns(table_or_model))
            if name == target_name
        )
    except EmptyIterableError:
        msg = f"Unable to map {series_name!r} to {table_or_model}"
        raise error(msg) from None
    _check_series_against_table_column(series, column)
    return column.name, _yield_insertion_elements(series)


class SeriesNameSnakeCaseNotInTableError(ValueError):
    """Raised when a Series name is not in a table, modulo snake case."""


class SeriesNameNotInTableError(ValueError):
    """Raised when a Series name is not in a table."""


@beartype
def _check_series_against_table_column(
    series: Series,
    table_column: Column[Any],
    /,
) -> None:
    """Check if a series can be inserted into a column."""
    py_type = table_column.type.python_type
    if not (
        (
            has_dtype(series, (bool, boolean))
            and issubclass(py_type, (bool, int))
        )
        or (has_dtype(series, float) and issubclass(py_type, float))
        or (
            has_dtype(series, datetime64ns)
            and issubclass(py_type, dt.date)
            and not issubclass(py_type, dt.datetime)
        )
        or (
            has_dtype(series, datetime64nsutc)
            and issubclass(py_type, dt.datetime)
        )
        or (has_dtype(series, (int, Int64)) and (py_type, int))
        or (has_dtype(series, string) and (py_type, str))
    ):
        msg = f"{series=}, {table_column=}"
        raise SeriesAgainstTableColumnError(msg)


class SeriesAgainstTableColumnError(TypeError):
    """Raised when a series has incompatible dtype with a table column."""


@beartype
def _yield_insertion_elements(series: Series, /) -> Iterator[Any]:
    """Yield the elements for insertion."""
    if has_dtype(series, (bool, boolean)):
        cast = bool
    elif has_dtype(series, datetime64ns):
        if (series.notna() & (series != series.dt.normalize())).any():
            raise DatesWithTimeComponentsError(str(series))
        cast = timestamp_to_date
    elif has_dtype(series, datetime64nsutc):
        cast = timestamp_to_datetime
    elif has_dtype(series, float):
        cast = float
    elif has_dtype(series, (int, Int64)):
        cast = int
    elif has_dtype(series, string):
        cast = str
    else:
        msg = f"Invalid dtype: {series=}"
        raise TypeError(msg)
    return (None if n else cast(v) for n, v in zip(series.isna(), series))


class DatesWithTimeComponentsError(ValueError):
    """Raised when dates with time components are encountered."""


@overload
def select_to_dataframe(
    sel: Select,
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    snake: bool = False,
    stream: int,
) -> Iterator[DataFrame]:
    ...


@overload
def select_to_dataframe(
    sel: Select,
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    snake: bool = False,
    stream: None = None,
) -> DataFrame:
    ...


@beartype
def select_to_dataframe(
    sel: Select,
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    snake: bool = False,
    stream: Optional[int] = None,
) -> Union[DataFrame, Iterator[DataFrame]]:
    """Read a table from a database into a DataFrame.

    Optionally stream it in chunks.
    """
    _check_select_for_duplicates(sel)
    if stream is None:
        with yield_connection(engine_or_conn) as conn:
            rows = conn.execute(sel).all()
            return _rows_to_dataframe(sel, rows, snake=snake)
    return _stream_dataframes(sel, engine_or_conn, stream, snake=snake)


@beartype
def _check_select_for_duplicates(sel: Select, /) -> None:
    """Check a select statement contains no duplicates."""
    col_names = [col.name for col in sel.selected_columns.values()]
    try:
        check_duplicates(col_names)
    except DuplicatedError:
        msg = f"{col_names=}"
        raise DuplicateColumnError(msg) from None


@beartype
def _rows_to_dataframe(
    sel: Select,
    rows: Iterable[Row],
    /,
    *,
    snake: bool = False,
) -> DataFrame:
    """Convert a set of rows into a DataFrame."""
    dtypes = {
        col.name: _table_column_to_dtype(col)
        for col in sel.selected_columns.values()
    }
    df = DataFrame(rows, columns=list(dtypes)).astype(dtypes)
    if snake:
        return _dataframe_columns_to_snake(df)
    return df


@beartype
def _table_column_to_dtype(column: ColumnElement[Any], /) -> Any:
    """Map a table column to a DataFrame dtype."""
    py_type = column.type.python_type
    if py_type is bool:
        return boolean
    if (py_type is float) or (py_type is Decimal):
        return float
    if py_type is int:
        return Int64
    if py_type is str:
        return string
    if issubclass(py_type, dt.date):
        return datetime64ns
    msg = f"Invalid type: {py_type=}"  # pragma: no cover
    raise TypeError(msg)  # pragma: no cover


@beartype
def _dataframe_columns_to_snake(df: DataFrame, /) -> DataFrame:
    """Convert the columns of a DataFrame to snake case."""
    columns = [c for c in df.columns if isinstance(c, str)]
    mapping = {col: snake_case(col) for col in columns}
    try:
        check_duplicates(mapping.values())
    except DuplicatedError:
        raise SnakeCaseDuplicateColumnsError(str(df)) from None
    return df.rename(columns=mapping)


class SnakeCaseDuplicateColumnsError(ValueError):
    """Raised when a DataFrame contains duplicated snake-case columns."""


@beartype
def _stream_dataframes(
    sel: Select,
    engine_or_conn: Union[Engine, Connection],
    stream: int,
    /,
    *,
    snake: bool = False,
) -> Iterator[DataFrame]:
    if stream <= 0:
        raise NonPositiveStreamError(str(stream))
    if isinstance(engine_or_conn, Engine):
        with engine_or_conn.begin() as conn:
            yield from _stream_dataframes(sel, conn, stream, snake=snake)
    else:
        for rows in (
            engine_or_conn.execution_options(yield_per=stream)
            .execute(sel)
            .partitions()
        ):
            yield _rows_to_dataframe(sel, rows, snake=snake)


class NonPositiveStreamError(ValueError):
    """Raised when the `stream` arguemnt is non-positive."""



from collections.abc import Iterable
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path
from smtplib import SMTP
from typing import Any, Optional

from beartype import beartype

from utilities.beartype import IterableStrs
from utilities.pathlib import PathLike


@beartype
def send_email(
    from_: str,
    to: IterableStrs,
    /,
    *,
    subject: Optional[str] = None,
    contents: Any = None,
    subtype: str = "plain",
    host: str = "",
    port: int = 0,
    attachments: Optional[Iterable[PathLike]] = None,
) -> None:
    """Send an email."""
    message = MIMEMultipart()
    message["From"] = from_
    message["To"] = ",".join(to)
    if subject is not None:
        message["Subject"] = subject
    if contents is not None:
        if isinstance(contents, str):
            text = MIMEText(contents, subtype)
        else:
            try:
                from airium import Airium
            except ModuleNotFoundError:  # pragma: no cover
                raise InvalidContentsError(contents) from None
            else:
                if not isinstance(contents, Airium):
                    raise InvalidContentsError(contents)
                text = MIMEText(str(contents), "html")
        message.attach(text)
    if attachments is not None:
        for attachment in attachments:
            _add_attachment(attachment, message)
    with SMTP(host=host, port=port) as smtp:
        _ = smtp.send_message(message)


@beartype
def _add_attachment(path: PathLike, message: MIMEMultipart, /) -> None:
    """Add an attachment to an email."""
    path = Path(path)
    name = path.name
    with path.open(mode="rb") as fh:
        part = MIMEApplication(fh.read(), Name=name)
    part["Content-Disposition"] = f"attachment; filename{name}"
    message.attach(part)


class InvalidContentsError(TypeError):
    """Raised when an invalid set of contents is encountered."""



from typing import Annotated, Any

from beartype.door import die_if_unbearable
from numpy import empty, zeros
from numpy.typing import NDArray
from pytest import mark, param

from utilities.beartype.numpy import (
    DTypeB,
    DTypeDD,
    DTypeDns,
    DTypeDY,
    DTypeF,
    DTypeI,
    DTypeO,
    NDim0,
    NDim1,
    NDim2,
    NDim3,
)
from utilities.numpy import datetime64D, datetime64ns, datetime64Y


class TestAnnotations:
    @mark.parametrize(
        ("dtype", "hint"),
        [
            param(bool, DTypeB),
            param(datetime64D, DTypeDD),
            param(datetime64Y, DTypeDY),
            param(datetime64ns, DTypeDns),
            param(float, DTypeF),
            param(int, DTypeI),
            param(object, DTypeO),
        ],
    )
    def test_dtypes(self, dtype: Any, hint: Any) -> None:
        arr = empty(0, dtype=dtype)
        die_if_unbearable(arr, Annotated[NDArray[Any], hint])

    @mark.parametrize(
        ("ndim", "hint"),
        [
            param(0, NDim0),
            param(1, NDim1),
            param(2, NDim2),
            param(3, NDim3),
        ],
    )
    def test_ndims(self, ndim: int, hint: Any) -> None:
        arr = empty(zeros(ndim, dtype=int), dtype=float)
        die_if_unbearable(arr, Annotated[NDArray[Any], hint])



import datetime as dt
from typing import Any, Union, cast

from beartype import beartype
from pandas import (
    DataFrame,
    DatetimeTZDtype,
    Index,
    NaT,
    RangeIndex,
    Series,
    Timestamp,
)

from utilities.datetime import UTC

Int64 = "Int64"
boolean = "boolean"
category = "category"
string = "string"
datetime64nsutc = DatetimeTZDtype(tz=UTC)


@beartype
def check_range_index(obj: Union[Index, Series, DataFrame], /) -> None:
    """Check if a RangeIndex is the default one."""
    if isinstance(obj, Index):
        if not isinstance(obj, RangeIndex):
            msg = f"Invalid type: {obj=}"
            raise TypeError(msg)
        if obj.start != 0:
            msg = f"{obj=}"
            raise RangeIndexStartError(msg)
        if obj.step != 1:
            msg = f"{obj=}"
            raise RangeIndexStepError(msg)
        if obj.name is not None:
            msg = f"{obj=}"
            raise RangeIndexNameError(msg)
    else:
        try:
            check_range_index(obj.index)
        except (
            TypeError,
            RangeIndexStartError,
            RangeIndexStepError,
            RangeIndexNameError,
        ) as error:
            msg = f"{obj=}"
            if isinstance(obj, Series):
                raise SeriesRangeIndexError(msg) from error
            raise DataFrameRangeIndexError(msg) from error


class RangeIndexStartError(ValueError):
    """Raised when a RangeIndex start is not 0."""


class RangeIndexStepError(ValueError):
    """Raised when a RangeIndex step is not 1."""


class RangeIndexNameError(ValueError):
    """Raised when a RangeIndex name is not None."""


class SeriesRangeIndexError(ValueError):
    """Raised when Series does not have a standard RangeIndex."""


class DataFrameRangeIndexError(ValueError):
    """Raised when DataFrame does not have a standard RangeIndex."""


@beartype
def timestamp_to_date(timestamp: Any, /) -> dt.date:
    """Convert a timestamp to a date."""
    return timestamp_to_datetime(timestamp).date()


@beartype
def timestamp_to_datetime(timestamp: Any, /) -> dt.datetime:
    """Convert a timestamp to a datetime."""
    if timestamp is NaT:
        msg = f"Invalid value: {timestamp}"
        raise ValueError(msg)
    datetime = cast(dt.datetime, timestamp.to_pydatetime())
    if datetime.tzinfo is None:
        return datetime.replace(tzinfo=UTC)
    return datetime


@beartype
def _timestamp_minmax_to_date(
    timestamp: Timestamp,
    method_name: str,
    /,
) -> dt.date:
    """Get the maximum Timestamp as a date."""
    method = getattr(timestamp, method_name)
    rounded = cast(Timestamp, method("D"))
    return timestamp_to_date(rounded)


TIMESTAMP_MIN_AS_DATE = _timestamp_minmax_to_date(Timestamp.min, "ceil")
TIMESTAMP_MAX_AS_DATE = _timestamp_minmax_to_date(Timestamp.max, "floor")


@beartype
def _timestamp_minmax_to_datetime(
    timestamp: Timestamp,
    method_name: str,
    /,
) -> dt.datetime:
    """Get the maximum Timestamp as a datetime."""
    method = getattr(timestamp, method_name)
    rounded = cast(Timestamp, method("us"))
    return timestamp_to_datetime(rounded)


TIMESTAMP_MIN_AS_DATETIME = _timestamp_minmax_to_datetime(Timestamp.min, "ceil")
TIMESTAMP_MAX_AS_DATETIME = _timestamp_minmax_to_datetime(
    Timestamp.max,
    "floor",
)



from typing import Annotated, Any, cast

from numpy import bool_, float64, int64, object_
from numpy.typing import NDArray

from utilities.beartype.numpy import NDim0, NDim1, NDim2, NDim3
from utilities.numpy import datetime64D, datetime64ns, datetime64Y

# dtype
NDArrayB = NDArray[bool_]
NDArrayDD = NDArray[cast(Any, datetime64D)]
NDArrayDY = NDArray[cast(Any, datetime64Y)]
NDArrayDns = NDArray[cast(Any, datetime64ns)]
NDArrayF = NDArray[float64]
NDArrayI = NDArray[int64]
NDArrayO = NDArray[object_]

# ndim
NDArray0 = Annotated[NDArray[Any], NDim0]
NDArray1 = Annotated[NDArray[Any], NDim1]
NDArray2 = Annotated[NDArray[Any], NDim2]
NDArray3 = Annotated[NDArray[Any], NDim3]

# compound
NDArrayB0 = Annotated[NDArrayB, NDim0]
NDArrayDD0 = Annotated[NDArrayDD, NDim0]
NDArrayDY0 = Annotated[NDArrayDY, NDim0]
NDArrayDns0 = Annotated[NDArrayDns, NDim0]
NDArrayF0 = Annotated[NDArrayF, NDim0]
NDArrayI0 = Annotated[NDArrayI, NDim0]
NDArrayO0 = Annotated[NDArrayO, NDim0]

NDArrayB1 = Annotated[NDArrayB, NDim1]
NDArrayDD1 = Annotated[NDArrayDD, NDim1]
NDArrayDY1 = Annotated[NDArrayDY, NDim1]
NDArrayDns1 = Annotated[NDArrayDns, NDim1]
NDArrayF1 = Annotated[NDArrayF, NDim1]
NDArrayI1 = Annotated[NDArrayI, NDim1]
NDArrayO1 = Annotated[NDArrayO, NDim1]

NDArrayB2 = Annotated[NDArrayB, NDim2]
NDArrayDD2 = Annotated[NDArrayDD, NDim2]
NDArrayDY2 = Annotated[NDArrayDY, NDim2]
NDArrayDns2 = Annotated[NDArrayDns, NDim2]
NDArrayF2 = Annotated[NDArrayF, NDim2]
NDArrayI2 = Annotated[NDArrayI, NDim2]
NDArrayO2 = Annotated[NDArrayO, NDim2]

NDArrayB3 = Annotated[NDArrayB, NDim3]
NDArrayDD3 = Annotated[NDArrayDD, NDim3]
NDArrayDY3 = Annotated[NDArrayDY, NDim3]
NDArrayDns3 = Annotated[NDArrayDns, NDim3]
NDArrayF3 = Annotated[NDArrayF, NDim3]
NDArrayI3 = Annotated[NDArrayI, NDim3]
NDArrayO3 = Annotated[NDArrayO, NDim3]



from math import inf, isfinite, isinf, isnan
from pathlib import Path
from re import search
from typing import Optional

from hypothesis import given
from hypothesis.errors import InvalidArgument
from hypothesis.strategies import (
    DataObject,
    DrawFn,
    booleans,
    composite,
    data,
    floats,
    integers,
    just,
    none,
    sets,
)
from more_itertools import pairwise
from pytest import mark, param, raises

from utilities.hypothesis import (
    assume_does_not_raise,
    floats_extra,
    lists_fixed_length,
    setup_hypothesis_profiles,
    slices,
    temp_dirs,
    temp_paths,
    text_ascii,
    text_clean,
    text_printable,
)
from utilities.tempfile import TemporaryDirectory


class TestAssumeDoesNotRaise:
    @given(x=booleans())
    def test_no_match_and_suppressed(self, x: bool) -> None:
        with assume_does_not_raise(ValueError):
            if x is True:
                msg = "x is True"
                raise ValueError(msg)
        assert x is False

    @given(x=booleans())
    def test_no_match_and_not_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with raises(ValueError, match=msg), assume_does_not_raise(
                RuntimeError,
            ):
                raise ValueError(msg)

    @given(x=booleans())
    def test_with_match_and_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with assume_does_not_raise(ValueError, match=msg):
                raise ValueError(msg)
        assert x is False

    @given(x=just(True))
    def test_with_match_and_not_suppressed(self, x: bool) -> None:
        msg = "x is True"
        if x is True:
            with raises(ValueError, match=msg), assume_does_not_raise(
                ValueError,
                match="wrong",
            ):
                raise ValueError(msg)


class TestFloatsExtra:
    @given(
        data=data(),
        min_value=floats() | none(),
        max_value=floats() | none(),
        allow_nan=booleans(),
        allow_inf=booleans(),
        allow_pos_inf=booleans(),
        allow_neg_inf=booleans(),
        integral=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        min_value: Optional[float],
        max_value: Optional[float],
        allow_nan: bool,
        allow_inf: bool,
        allow_pos_inf: bool,
        allow_neg_inf: bool,
        integral: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument):
            x = data.draw(
                floats_extra(
                    min_value=min_value,
                    max_value=max_value,
                    allow_nan=allow_nan,
                    allow_inf=allow_inf,
                    allow_pos_inf=allow_pos_inf,
                    allow_neg_inf=allow_neg_inf,
                    integral=integral,
                ),
            )
        if min_value is not None:
            assert (isfinite(x) and x >= min_value) or not isfinite(x)
        if max_value is not None:
            assert (isfinite(x) and x <= max_value) or not isfinite(x)
        if not allow_nan:
            assert not isnan(x)
        if not allow_inf:
            if not (allow_pos_inf or allow_neg_inf):
                assert not isinf(x)
            if not allow_pos_inf:
                assert x != inf
            if not allow_neg_inf:
                assert x != -inf
        if integral:
            assert (isfinite(x) and x == round(x)) or not isfinite(x)


class TestLiftDraw:
    @given(data=data(), x=booleans())
    def test_fixed(self, data: DataObject, x: bool) -> None:
        @composite
        def func(_draw: DrawFn, /) -> bool:
            _ = _draw
            return x

        result = data.draw(func())
        assert result is x

    @given(data=data())
    def test_strategy(self, data: DataObject) -> None:
        @composite
        def func(_draw: DrawFn, /) -> bool:
            return _draw(booleans())

        result = data.draw(func())
        assert isinstance(result, bool)


class TestListsFixedLength:
    @given(data=data(), size=integers(1, 10))
    @mark.parametrize(
        "unique",
        [param(True, id="unique"), param(False, id="no unique")],
    )
    @mark.parametrize(
        "sorted_",
        [param(True, id="sorted"), param(False, id="no sorted")],
    )
    def test_main(
        self,
        data: DataObject,
        size: int,
        unique: bool,
        sorted_: bool,
    ) -> None:
        result = data.draw(
            lists_fixed_length(integers(), size, unique=unique, sorted=sorted_),
        )
        assert isinstance(result, list)
        assert len(result) == size
        if unique:
            assert len(set(result)) == len(result)
        if sorted_:
            assert sorted(result) == result


class TestSlices:
    @given(data=data(), iter_len=integers(0, 10))
    def test_main(self, data: DataObject, iter_len: int) -> None:
        slice_len = data.draw(integers(0, iter_len) | none())
        slice_ = data.draw(slices(iter_len, slice_len=slice_len))
        range_slice = range(iter_len)[slice_]
        assert all(i + 1 == j for i, j in pairwise(range_slice))
        if slice_len is not None:
            assert len(range_slice) == slice_len

    @given(data=data(), iter_len=integers(0, 10))
    def test_error(self, data: DataObject, iter_len: int) -> None:
        with raises(
            InvalidArgument,
            match=r"Slice length \d+ exceeds iterable length \d+",
        ):
            _ = data.draw(slices(iter_len, slice_len=iter_len + 1))


class TestSetupHypothesisProfiles:
    def test_main(self) -> None:
        setup_hypothesis_profiles()


class TestTempDirs:
    @given(temp_dir=temp_dirs())
    def test_main(self, temp_dir: TemporaryDirectory) -> None:
        _test_temp_path(temp_dir.name)

    @given(
        temp_dir=temp_dirs(),
        contents=sets(text_ascii(min_size=1), max_size=10),
    )
    def test_writing_files(
        self,
        temp_dir: TemporaryDirectory,
        contents: set[str],
    ) -> None:
        _test_writing_to_temp_path(temp_dir.name, contents)


class TestTempPaths:
    @given(temp_path=temp_paths())
    def test_main(self, temp_path: Path) -> None:
        _test_temp_path(temp_path)

    @given(
        temp_path=temp_paths(),
        contents=sets(text_ascii(min_size=1), max_size=10),
    )
    def test_writing_files(self, temp_path: Path, contents: set[str]) -> None:
        _test_writing_to_temp_path(temp_path, contents)


def _test_temp_path(path: Path, /) -> None:
    assert path.is_dir()
    assert len(set(path.iterdir())) == 0


def _test_writing_to_temp_path(path: Path, contents: set[str], /) -> None:
    assert len(set(path.iterdir())) == 0
    for content in contents:
        path.joinpath(content).touch()
    assert len(set(path.iterdir())) == len(contents)


class TestTextAscii:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: Optional[int],
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_ascii(
                    min_size=min_size,
                    max_size=max_size,
                    disallow_na=disallow_na,
                ),
            )
        assert search("^[A-Za-z]*$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"


class TestTextClean:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: Optional[int],
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_clean(
                    min_size=min_size,
                    max_size=max_size,
                    disallow_na=disallow_na,
                ),
            )
        assert search("^\\S[^\\r\\n]*$|^$", text)
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"


class TestTextPrintable:
    @given(
        data=data(),
        min_size=integers(0, 100),
        max_size=integers(0, 100) | none(),
        disallow_na=booleans(),
    )
    def test_main(
        self,
        data: DataObject,
        min_size: int,
        max_size: Optional[int],
        disallow_na: bool,
    ) -> None:
        with assume_does_not_raise(InvalidArgument, AssertionError):
            text = data.draw(
                text_printable(
                    min_size=min_size,
                    max_size=max_size,
                    disallow_na=disallow_na,
                ),
            )
        assert search(
            r"^[0-9A-Za-z!\"#$%&'()*+,-./:;<=>?@\[\\\]^_`{|}~\s]*$",
            text,
        )
        assert len(text) >= min_size
        if max_size is not None:
            assert len(text) <= max_size
        if disallow_na:
            assert text != "NA"



from pytest import mark, param, raises

from utilities.fastapi import APIRouter


class TestAPIRouter:
    @mark.parametrize("route", [param("/"), param("/home")])
    def test_main(self, route: str) -> None:
        router = APIRouter()

        @router.get(route)
        def _() -> None:
            return None

    def test_error(self) -> None:
        router = APIRouter()

        with raises(ValueError, match="Invalid route"):

            @router.get("/home/")
            def _() -> None:
                return None



import datetime as dt
from enum import Enum as _Enum
from typing import Any, Generic, Optional, TypeVar

from beartype import beartype
from click import Context, Parameter, ParamType, option

from utilities.datetime import (
    ParseDateError,
    ParseDateTimeError,
    ParseTimeError,
    TimedeltaError,
    ensure_date,
    ensure_datetime,
    ensure_time,
    ensure_timedelta,
)
from utilities.enum import (
    MultipleMatchingMembersError,
    NoMatchingMemberError,
    ensure_enum,
)
from utilities.logging import LogLevel


class Date(ParamType):
    """A date-valued parameter."""

    name = "date"

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> dt.date:
        """Convert a value into the `Date` type."""
        try:
            return ensure_date(value)
        except ParseDateError:
            self.fail(f"Unable to parse {value}", param, ctx)


class DateTime(ParamType):
    """A datetime-valued parameter."""

    name = "datetime"

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> dt.date:
        """Convert a value into the `DateTime` type."""
        try:
            return ensure_datetime(value)
        except ParseDateTimeError:
            self.fail(f"Unable to parse {value}", param, ctx)


class Time(ParamType):
    """A time-valued parameter."""

    name = "time"

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> dt.time:
        """Convert a value into the `Time` type."""
        try:
            return ensure_time(value)
        except ParseTimeError:
            self.fail(f"Unable to parse {value}", param, ctx)


class Timedelta(ParamType):
    """A timedelta-valued parameter."""

    name = "timedelta"

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> dt.timedelta:
        """Convert a value into the `Timedelta` type."""
        try:
            return ensure_timedelta(value)
        except TimedeltaError:
            self.fail(f"Unable to parse {value}", param, ctx)


_E = TypeVar("_E", bound=_Enum)


class Enum(ParamType, Generic[_E]):
    """An enum-valued parameter."""

    name = "enum"

    @beartype
    def __init__(
        self,
        enum: type[_E],
        /,
        *,
        case_sensitive: bool = True,
    ) -> None:
        super().__init__()
        self._enum = enum
        self._case_sensitive = case_sensitive

    @beartype
    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> _E:
        """Convert a value into the `Enum` type."""
        try:
            return ensure_enum(
                self._enum,
                value,
                case_sensitive=self._case_sensitive,
            )
        except (NoMatchingMemberError, MultipleMatchingMembersError):
            return self.fail(f"Unable to parse {value}", param, ctx)


log_level_option = option(
    "-ll",
    "--log-level",
    type=Enum(LogLevel, case_sensitive=False),
    default=LogLevel.INFO,
    show_default=True,
    help="The logging level",
)



from collections.abc import Iterable
from os import environ

from _pytest.config import Config
from _pytest.config.argparsing import Parser
from _pytest.python import Function
from beartype import beartype
from pytest import mark

from utilities.beartype import IterableStrs


@beartype
def add_pytest_addoption(parser: Parser, options: IterableStrs, /) -> None:
    """Add the `--slow`, etc options to pytest.

    Usage:

        def pytest_addoption(parser):
            add_pytest_addoption(parser, ["slow"])
    """
    for opt in options:
        parser.addoption(
            f"--{opt}",
            action="store_true",
            default=False,
            help=f"run tests marked {opt!r}",
        )


@beartype
def add_pytest_collection_modifyitems(
    config: Config,
    items: Iterable[Function],
    options: IterableStrs,
    /,
) -> None:
    """Add the @mark.skips as necessary.

    Usage:

        def pytest_collection_modifyitems(config, items):
            add_pytest_collection_modifyitems(config, items, ["slow"])
    """
    options = list(options)
    missing = {opt for opt in options if not config.getoption(f"--{opt}")}
    for item in items:
        opts_on_item = [opt for opt in options if opt in item.keywords]
        if len(missing & set(opts_on_item)) >= 1:
            flags = [f"--{opt}" for opt in opts_on_item]
            joined = " ".join(flags)
            item.add_marker(mark.skip(reason=f"pass {joined}"))


@beartype
def add_pytest_configure(
    config: Config,
    options: Iterable[tuple[str, str]],
    /,
) -> None:
    """Add the `--slow`, etc markers to pytest.

    Usage:
        def pytest_configure(config):
            add_pytest_configure(config, [("slow", "slow to run")])
    """
    for opt, desc in options:
        config.addinivalue_line("markers", f"{opt}: mark test as {desc}")


@beartype
def is_pytest() -> bool:
    """Check if pytest is currently running."""
    return "PYTEST_CURRENT_TEST" in environ



from collections.abc import Callable, Iterable, Iterator
from contextlib import contextmanager, suppress
from functools import reduce
from operator import ge, le
from typing import Any, Literal, NoReturn, Optional, Union, cast

from beartype import beartype
from more_itertools import chunked
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Enum,
    Float,
    Interval,
    LargeBinary,
    MetaData,
    Numeric,
    Select,
    String,
    Table,
    Unicode,
    UnicodeText,
    Uuid,
    and_,
    case,
    quoted_name,
    text,
)
from sqlalchemy import create_engine as _create_engine
from sqlalchemy.dialects.mssql import dialect as mssql_dialect
from sqlalchemy.dialects.mysql import dialect as mysql_dialect
from sqlalchemy.dialects.oracle import dialect as oracle_dialect
from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.engine import URL, Connection, Engine
from sqlalchemy.exc import (
    DatabaseError,
    NoSuchTableError,
)
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.pool import NullPool, Pool
from sqlalchemy.sql.base import ReadOnlyColumnCollection

from utilities.errors import redirect_error
from utilities.inflection import snake_case
from utilities.more_itertools import one
from utilities.typing import never


@beartype
def check_table_against_reflection(
    table_or_model: Any,
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    schema: Optional[str] = None,
) -> None:
    """Check that a table equals its reflection."""
    reflected = _reflect_table(table_or_model, engine_or_conn, schema=schema)
    check_tables_equal(reflected, table_or_model)


@beartype
def _reflect_table(
    table_or_model: Any,
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    schema: Optional[str] = None,
) -> Table:
    """Reflect a table from a database."""
    name = get_table_name(table_or_model)
    metadata = MetaData(schema=schema)
    with yield_connection(engine_or_conn) as conn:
        return Table(name, metadata, autoload_with=conn)


@beartype
def check_tables_equal(
    x: Any,
    y: Any,
    /,
    *,
    snake_table: bool = False,
    snake_columns: bool = False,
) -> None:
    """Check that a pair of tables are equal."""
    x_t, y_t = map(get_table, [x, y])
    _check_table_or_column_names_equal(x_t.name, y_t.name, snake=snake_table)
    _check_column_collections_equal(
        x_t.columns,
        y_t.columns,
        snake=snake_columns,
    )


@beartype
def _check_table_or_column_names_equal(
    x: Union[str, quoted_name],
    y: Union[str, quoted_name],
    /,
    *,
    snake: bool = False,
) -> None:
    """Check that a pair of table/columns' names are equal."""
    x, y = (str(i) if isinstance(i, quoted_name) else i for i in [x, y])
    msg = f"{x=}, {y=}"
    if snake and (snake_case(x) != snake_case(y)):
        raise UnequalTableOrColumnSnakeCaseNamesError(msg)
    if (not snake) and (x != y):
        raise UnequalTableOrColumnNamesError(msg)


class UnequalTableOrColumnNamesError(ValueError):
    """Raised when two table/columns' names differ."""


class UnequalTableOrColumnSnakeCaseNamesError(ValueError):
    """Raised when two table/columns' snake case names differ."""


@beartype
def _check_column_collections_equal(
    x: ReadOnlyColumnCollection[Any, Any],
    y: ReadOnlyColumnCollection[Any, Any],
    /,
    *,
    snake: bool = False,
) -> None:
    """Check that a pair of column collections are equal."""
    msg = f"{x=}, {y=}"
    if len(x) != len(y):
        raise UnequalNumberOfColumnsError(msg)
    for x_i, y_i in zip(x, y):
        _check_columns_equal(x_i, y_i, snake=snake)


class UnequalNumberOfColumnsError(ValueError):
    """Raised when two column collections' lengths differ."""


@beartype
def _check_columns_equal(
    x: Column[Any],
    y: Column[Any],
    /,
    *,
    snake: bool = False,
) -> None:
    """Check that a pair of columns are equal."""
    _check_table_or_column_names_equal(x.name, y.name, snake=snake)
    _check_column_types_equal(x.type, y.type)
    if x.primary_key != y.primary_key:
        msg = f"{x.primary_key=}, {y.primary_key=}"
        raise UnequalPrimaryKeyStatusError(msg)
    if x.nullable != y.nullable:
        msg = f"{x.nullable=}, {y.nullable=}"
        raise UnequalNullableStatusError(msg)


class UnequalPrimaryKeyStatusError(ValueError):
    """Raised when two columns differ in primary key status."""


class UnequalNullableStatusError(ValueError):
    """Raised when two columns differ in nullable status."""


@beartype
def _check_column_types_equal(  # noqa: C901, PLR0912, PLR0915
    x: Any,
    y: Any,
    /,
) -> None:
    """Check that a pair of column types are equal."""
    x_inst, y_inst = (i() if isinstance(i, type) else i for i in [x, y])
    x_cls, y_cls = (i._type_affinity for i in [x_inst, y_inst])  # noqa: SLF001
    msg = f"{x=}, {y=}"
    if not (isinstance(x_inst, y_cls) and isinstance(y_inst, x_cls)):
        raise UnequalColumnTypesError(msg)
    if isinstance(x_inst, Boolean) and isinstance(y_inst, Boolean):
        if x_inst.create_constraint is not y_inst.create_constraint:
            raise UnequalBooleanColumnCreateConstraintError(msg)
        if x_inst.name != y_inst.name:
            raise UnequalBooleanColumnNameError(msg)
    if isinstance(x_inst, Enum) and isinstance(y_inst, Enum):
        x_enum, y_enum = (cast(Any, i).enum_class for i in [x_inst, y_inst])
        if ((x_enum is None) and (y_enum is not None)) or (
            (x_enum is not None)
            and (y_enum is None)
            or (
                (x_enum is not None)
                and (y_enum is not None)
                and not (
                    issubclass(x_enum, y_enum) and issubclass(y_enum, x_enum)
                )
            )
        ):
            raise UnequalEnumColumnTypesError(msg)
        if x_inst.create_constraint is not y_inst.create_constraint:
            raise UnequalEnumColumnCreateConstraintError(msg)
        if x_inst.native_enum is not y_inst.native_enum:
            raise UnequalEnumColumnNativeEnumError(msg)
        if x_inst.length != y_inst.length:
            raise UnequalEnumColumnLengthError(msg)
        if x_inst.inherit_schema is not y_inst.inherit_schema:
            raise UnequalEnumColumnInheritSchemaError(msg)
    if (
        isinstance(x_inst, (Float, Numeric))
        and isinstance(y_inst, (Float, Numeric))
        and (x_inst.asdecimal is not y_inst.asdecimal)
    ):
        raise UnequalFloatColumnAsDecimalError(msg)
    if (
        isinstance(x_inst, DateTime)
        and isinstance(y_inst, DateTime)
        and (x_inst.timezone is not y_inst.timezone)
    ):
        raise UnequalDateTimeColumnTimezoneError(msg)
    if isinstance(x_inst, (Float, Numeric)) and isinstance(
        y_inst,
        (Float, Numeric),
    ):
        if x_inst.precision != y_inst.precision:
            raise UnequalFloatColumnPrecisionsError(msg)
        if x_inst.decimal_return_scale != y_inst.decimal_return_scale:
            raise UnequalFloatColumnDecimalReturnScaleError(msg)
        if x_inst.scale != y_inst.scale:
            raise UnequalNumericScaleError(msg)
    if isinstance(x_inst, Interval) and isinstance(y_inst, Interval):
        if x_inst.native is not y_inst.native:
            raise UnequalIntervalColumnNativeError(msg)
        if x_inst.second_precision != y_inst.second_precision:
            raise UnequalIntervalColumnSecondPrecisionError(msg)
        if x_inst.day_precision != y_inst.day_precision:
            raise UnequalIntervalColumnDayPrecisionError(msg)
    if (
        isinstance(x_inst, LargeBinary)
        and isinstance(y_inst, LargeBinary)
        and (x_inst.length != y_inst.length)
    ):
        raise UnequalLargeBinaryColumnLengthError(msg)
    if isinstance(x_inst, (String, Unicode, UnicodeText)) and isinstance(
        y_inst,
        (String, Unicode, UnicodeText),
    ):
        if x_inst.length != y_inst.length:
            raise UnequalStringLengthError(msg)
        if x_inst.collation != y_inst.collation:
            raise UnequalStringCollationError(msg)
    if isinstance(x_inst, Uuid) and isinstance(y_inst, Uuid):
        if x_inst.as_uuid is not y_inst.as_uuid:
            raise UnequalUUIDAsUUIDError(msg)
        if x_inst.native_uuid is not y_inst.native_uuid:
            raise UnequalUUIDNativeUUIDError(msg)


class UnequalColumnTypesError(TypeError):
    """Raised when two columns' types differ."""


class UnequalBooleanColumnCreateConstraintError(TypeError):
    """Raised when two boolean columns' create constraints differ."""


class UnequalBooleanColumnNameError(TypeError):
    """Raised when two boolean columns' names differ."""


class UnequalDateTimeColumnTimezoneError(TypeError):
    """Raised when two datetime columns' timezones differ."""


class UnequalEnumColumnTypesError(TypeError):
    """Raised when two enum columns' types differ."""


class UnequalEnumColumnCreateConstraintError(TypeError):
    """Raised when two enum columns' create constraints differ."""


class UnequalEnumColumnNativeEnumError(TypeError):
    """Raised when two enum columns' native enums differ."""


class UnequalEnumColumnLengthError(TypeError):
    """Raised when two enum columns' lengths differ."""


class UnequalEnumColumnInheritSchemaError(TypeError):
    """Raised when two enum columns' inherit schemas differ."""


class UnequalFloatColumnPrecisionsError(TypeError):
    """Raised when two float columns' precisions differ."""


class UnequalFloatColumnAsDecimalError(TypeError):
    """Raised when two float columns' asdecimal differ."""


class UnequalFloatColumnDecimalReturnScaleError(TypeError):
    """Raised when two float columns' decimal return scales differ."""


class UnequalIntervalColumnNativeError(TypeError):
    """Raised when two intervals columns' native differ."""


class UnequalIntervalColumnSecondPrecisionError(TypeError):
    """Raised when two intervals columns' second precisions differ."""


class UnequalIntervalColumnDayPrecisionError(TypeError):
    """Raised when two intervals columns' day precisions differ."""


class UnequalLargeBinaryColumnLengthError(TypeError):
    """Raised when two large binary columns' lengths differ."""


class UnequalNumericScaleError(TypeError):
    """Raised when two numeric columns' scales differ."""


class UnequalStringLengthError(TypeError):
    """Raised when two string columns' lengths differ."""


class UnequalStringCollationError(TypeError):
    """Raised when two string columns' collations differ."""


class UnequalUUIDAsUUIDError(TypeError):
    """Raised when two UUID columns' as_uuid differ."""


class UnequalUUIDNativeUUIDError(TypeError):
    """Raised when two UUID columns' native UUID differ."""


@beartype
def check_engine(
    engine: Engine,
    /,
    *,
    num_tables: Optional[int] = None,
    num_columns: Optional[int] = None,
) -> None:
    """Check that an engine can connect.

    Optionally query for the number of tables, or the number of columns in
    such a table.
    """
    dialect = get_dialect(engine)
    if (  # pragma: no cover
        (dialect == "mssql")
        or (dialect == "mysql")
        or (dialect == "postgresql")
    ):
        query = "select * from information_schema.tables"  # pragma: no cover
    elif dialect == "oracle":  # pragma: no cover
        query = "select * from all_objects"
    elif dialect == "sqlite":
        query = "select * from sqlite_master where type='table'"
    else:
        return never(dialect)  # pragma: no cover
    with engine.begin() as conn:
        rows = conn.execute(text(query)).all()
    if (num_tables is not None) and (len(rows) != num_tables):
        msg = f"{len(rows)=}, {num_tables=}"
        raise IncorrectNumberOfTablesError(msg)
    if num_columns is not None:
        if len(rows) == 0:
            msg = f"{engine=}"
            raise NoTablesError(msg)
        if len(rows[0]) != num_columns:
            msg = f"{len(rows[0])=}, {num_columns=}"
            raise IncorrectNumberOfColumnsError(msg)
    return None


class IncorrectNumberOfTablesError(ValueError):
    """Raised when there are an incorrect number of tables."""


class NoTablesError(ValueError):
    """Raised when there are no tables."""


class IncorrectNumberOfColumnsError(ValueError):
    """Raised when there are an incorrect number of columns."""


@beartype
def columnwise_max(*columns: Any) -> Any:
    """Compute the columnwise max of a number of columns."""
    return _columnwise_minmax(*columns, op=ge)


@beartype
def columnwise_min(*columns: Any) -> Any:
    """Compute the columnwise min of a number of columns."""
    return _columnwise_minmax(*columns, op=le)


@beartype
def _columnwise_minmax(*columns: Any, op: Callable[[Any, Any], Any]) -> Any:
    """Compute the columnwise min of a number of columns."""

    @beartype
    def func(x: Any, y: Any, /) -> Any:
        x_none = x.is_(None)
        y_none = y.is_(None)
        col = case(
            (and_(x_none, y_none), None),
            (and_(~x_none, y_none), x),
            (and_(x_none, ~y_none), y),
            (op(x, y), x),
            else_=y,
        )
        # try auto-label
        names = {
            value
            for col in [x, y]
            if (value := getattr(col, "name", None)) is not None
        }
        try:
            (name,) = names
        except ValueError:
            return col
        else:
            return col.label(name)

    return reduce(func, columns)


@beartype
def create_engine(
    drivername: str,
    /,
    *,
    username: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    database: Optional[str] = None,
    poolclass: Optional[type[Pool]] = NullPool,
) -> Engine:
    """Create a SQLAlchemy engine."""
    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
    )
    return _create_engine(url, future=True, poolclass=poolclass)


Dialect = Literal["mssql", "mysql", "oracle", "postgresql", "sqlite"]


@beartype
def get_dialect(engine_or_conn: Union[Engine, Connection], /) -> Dialect:
    """Get the dialect of a database."""
    if isinstance(dialect := engine_or_conn.dialect, mssql_dialect):
        return "mssql"
    if isinstance(dialect, mysql_dialect):
        return "mysql"
    if isinstance(dialect, oracle_dialect):
        return "oracle"
    if isinstance(dialect, postgresql_dialect):
        return "postgresql"
    if isinstance(dialect, sqlite_dialect):
        return "sqlite"
    msg = f"{dialect=}"  # pragma: no cover
    raise UnsupportedDialectError(msg)  # pragma: no cover


class UnsupportedDialectError(TypeError):
    """Raised when a dialect is unsupported."""


@beartype
def ensure_table_created(
    table_or_model: Any,
    engine_or_connection: Union[Engine, Connection],
    /,
) -> None:
    """Ensure a table is created."""
    table = get_table(table_or_model)
    try:
        with yield_connection(engine_or_connection) as conn:
            table.create(conn)
    except DatabaseError as error:
        with suppress(TableAlreadyExistsError):
            redirect_to_table_already_exists_error(engine_or_connection, error)


@beartype
def ensure_table_dropped(
    table_or_model: Any,
    engine_or_conn: Union[Engine, Connection],
    /,
) -> None:
    """Ensure a table is dropped."""
    table = get_table(table_or_model)
    try:
        with yield_connection(engine_or_conn) as conn:
            table.drop(conn)
    except DatabaseError as error:
        with suppress(NoSuchTableError):
            redirect_to_no_such_table_error(engine_or_conn, error)


@beartype
def get_column_names(table_or_model: Any, /) -> list[str]:
    """Get the column names from a table or model."""
    return [col.name for col in get_columns(table_or_model)]


@beartype
def get_columns(table_or_model: Any, /) -> list[Column[Any]]:
    """Get the columns from a table or model."""
    return list(get_table(table_or_model).columns)


@beartype
def get_table(table_or_model: Any, /) -> Table:
    """Get the table from a ORM model."""
    if isinstance(table_or_model, Table):
        return table_or_model
    return table_or_model.__table__


@beartype
def get_table_name(table_or_model: Any, /) -> str:
    """Get the table name from a ORM model."""
    return get_table(table_or_model).name


@beartype
def model_to_dict(obj: Any, /) -> dict[str, Any]:
    """Construct a dictionary of elements for insertion."""
    cls = type(obj)

    @beartype
    def is_attr(attr: str, key: str, /) -> Optional[str]:
        if isinstance(value := getattr(cls, attr), InstrumentedAttribute) and (
            value.name == key
        ):
            return attr
        return None

    @beartype
    def yield_items() -> Iterator[tuple[str, Any]]:
        for key in get_column_names(cls):
            attr = one(
                attr for attr in dir(cls) if is_attr(attr, key) is not None
            )
            yield key, getattr(obj, attr)

    return dict(yield_items())


@beartype
def redirect_to_no_such_table_error(
    engine_or_conn: Union[Engine, Connection],
    error: DatabaseError,
    /,
) -> NoReturn:
    """Redirect to the `NoSuchTableError`."""
    dialect = get_dialect(engine_or_conn)
    if (  # pragma: no cover
        dialect == "mssql" or dialect == "mysql" or dialect == "postgresql"
    ):
        raise NotImplementedError(dialect)  # pragma: no cover
    if dialect == "oracle":  # pragma: no cover
        pattern = "ORA-00942: table or view does not exist"
    elif dialect == "sqlite":
        pattern = "no such table"
    else:
        return never(dialect)  # pragma: no cover
    return redirect_error(error, pattern, NoSuchTableError)


@beartype
def redirect_to_table_already_exists_error(
    engine_or_conn: Union[Engine, Connection],
    error: DatabaseError,
    /,
) -> NoReturn:
    """Redirect to the `TableAlreadyExistsError`."""
    dialect = get_dialect(engine_or_conn)
    if (  # pragma: no cover
        dialect == "mssql" or dialect == "mysql" or dialect == "postgresql"
    ):
        raise NotImplementedError(dialect)  # pragma: no cover
    if dialect == "oracle":  # pragma: no cover
        pattern = "ORA-00955: name is already used by an existing object"
    elif dialect == "sqlite":
        pattern = "table .* already exists"
    else:
        return never(dialect)  # pragma: no cover
    return redirect_error(error, pattern, TableAlreadyExistsError)


class TableAlreadyExistsError(Exception):
    """Raised when a table already exists."""


@contextmanager
@beartype
def yield_connection(
    engine_or_conn: Union[Engine, Connection],
    /,
) -> Iterator[Connection]:
    """Yield a connection."""
    if isinstance(engine_or_conn, Engine):
        with engine_or_conn.begin() as conn:
            yield conn
    else:
        yield engine_or_conn


@beartype
def yield_in_clause_rows(
    sel: Select,
    column: Any,
    values: Iterable[Any],
    engine_or_conn: Union[Engine, Connection],
    /,
    *,
    chunk_size: Optional[int] = None,
    frac: float = 0.95,
) -> Iterator[Any]:
    """Yield the rows from an `in` clause."""
    if chunk_size is None:
        dialect = get_dialect(engine_or_conn)
        if dialect == "mssql":  # pragma: no cover
            max_params = 2100
        elif dialect == "mysql":  # pragma: no cover
            max_params = 65535
        elif dialect == "oracle":  # pragma: no cover
            max_params = 1000
        elif dialect == "postgresql":  # pragma: no cover
            max_params = 32767
        elif dialect == "sqlite":
            max_params = 100
        else:
            return never(dialect)  # pragma: no cover
        chunk_size_use = round(frac * max_params)
    else:
        chunk_size_use = chunk_size
    with yield_connection(engine_or_conn) as conn:
        for values_i in chunked(values, chunk_size_use):
            sel_i = sel.where(column.in_(values_i))
            yield from conn.execute(sel_i).all()
    return None



from typing import Annotated

from xarray import DataArray

from utilities.beartype.numpy import (
    DTypeB,
    DTypeDns,
    DTypeF,
    DTypeI,
    DTypeO,
    NDim0,
    NDim1,
    NDim2,
    NDim3,
)

# dtype
DataArrayB = Annotated[DataArray, DTypeB]
DataArrayDns = Annotated[DataArray, DTypeDns]
DataArrayF = Annotated[DataArray, DTypeF]
DataArrayI = Annotated[DataArray, DTypeI]
DataArrayO = Annotated[DataArray, DTypeO]

# ndim
DataArray0 = Annotated[DataArray, NDim0]
DataArray1 = Annotated[DataArray, NDim1]
DataArray2 = Annotated[DataArray, NDim2]
DataArray3 = Annotated[DataArray, NDim3]

# compound
DataArrayB0 = Annotated[DataArray, DTypeB & NDim0]
DataArrayDns0 = Annotated[DataArray, DTypeDns & NDim0]
DataArrayF0 = Annotated[DataArray, DTypeF & NDim0]
DataArrayI0 = Annotated[DataArray, DTypeI & NDim0]
DataArrayO0 = Annotated[DataArray, DTypeO & NDim0]

DataArrayB1 = Annotated[DataArray, DTypeB & NDim1]
DataArrayDns1 = Annotated[DataArray, DTypeDns & NDim1]
DataArrayF1 = Annotated[DataArray, DTypeF & NDim1]
DataArrayI1 = Annotated[DataArray, DTypeI & NDim1]
DataArrayO1 = Annotated[DataArray, DTypeO & NDim1]

DataArrayB2 = Annotated[DataArray, DTypeB & NDim2]
DataArrayDns2 = Annotated[DataArray, DTypeDns & NDim2]
DataArrayF2 = Annotated[DataArray, DTypeF & NDim2]
DataArrayI2 = Annotated[DataArray, DTypeI & NDim2]
DataArrayO2 = Annotated[DataArray, DTypeO & NDim2]

DataArrayB3 = Annotated[DataArray, DTypeB & NDim3]
DataArrayDns3 = Annotated[DataArray, DTypeDns & NDim3]
DataArrayF3 = Annotated[DataArray, DTypeF & NDim3]
DataArrayI3 = Annotated[DataArray, DTypeI & NDim3]
DataArrayO3 = Annotated[DataArray, DTypeO & NDim3]



from typing import Any, Optional

from beartype import beartype
from hypothesis.extra.numpy import array_shapes, arrays, from_dtype
from hypothesis.strategies import (
    SearchStrategy,
    booleans,
    composite,
    integers,
    nothing,
)
from numpy import bool_, dtype, float64, iinfo, int64
from numpy.typing import NDArray

from utilities.hypothesis import floats_extra, lift_draw
from utilities.hypothesis.typing import MaybeSearchStrategy, Shape

_ARRAY_SHAPES = array_shapes()


@composite
@beartype
def bool_arrays(
    _draw: Any,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = _ARRAY_SHAPES,
    fill: Optional[SearchStrategy[Any]] = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArray[bool_]:
    """Strategy for generating arrays of booleans."""
    draw = lift_draw(_draw)
    return draw(
        arrays(
            bool,
            draw(shape),
            elements=booleans(),
            fill=fill,
            unique=draw(unique),
        ),
    )


@composite
@beartype
def float_arrays(
    _draw: Any,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = _ARRAY_SHAPES,
    min_value: MaybeSearchStrategy[Optional[float]] = None,
    max_value: MaybeSearchStrategy[Optional[float]] = None,
    allow_nan: MaybeSearchStrategy[bool] = False,
    allow_inf: MaybeSearchStrategy[bool] = False,
    allow_pos_inf: MaybeSearchStrategy[bool] = False,
    allow_neg_inf: MaybeSearchStrategy[bool] = False,
    integral: MaybeSearchStrategy[bool] = False,
    fill: Optional[SearchStrategy[Any]] = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArray[float64]:
    """Strategy for generating arrays of floats."""
    draw = lift_draw(_draw)
    elements = floats_extra(
        min_value=min_value,
        max_value=max_value,
        allow_nan=allow_nan,
        allow_inf=allow_inf,
        allow_pos_inf=allow_pos_inf,
        allow_neg_inf=allow_neg_inf,
        integral=integral,
    )
    return draw(
        arrays(
            float,
            draw(shape),
            elements=elements,
            fill=fill,
            unique=draw(unique),
        ),
    )


@composite
@beartype
def int_arrays(
    _draw: Any,
    /,
    *,
    shape: MaybeSearchStrategy[Shape] = _ARRAY_SHAPES,
    min_value: MaybeSearchStrategy[Optional[int]] = None,
    max_value: MaybeSearchStrategy[Optional[int]] = None,
    unique: MaybeSearchStrategy[bool] = False,
) -> NDArray[int64]:
    """Strategy for generating arrays of ints."""
    draw = lift_draw(_draw)
    info = iinfo(int64)
    min_value_, max_value_ = draw(min_value), draw(max_value)
    min_value_use = info.min if min_value_ is None else min_value_
    max_value_use = info.max if max_value_ is None else max_value_
    elements = integers(min_value=min_value_use, max_value=max_value_use)
    return draw(
        arrays(
            int,
            draw(shape),
            elements=elements,
            fill=nothing(),
            unique=draw(unique),
        ),
    )


@beartype
def int64s() -> SearchStrategy[int]:
    """Strategy for generating int64s."""
    return from_dtype(dtype(int64)).map(int)



from beartype.vale import IsAttr, IsEqual

from utilities.pandas import Int64, boolean, string

DTypeBoolean = IsAttr["dtype", IsEqual[boolean]]
DTypeInt64 = IsAttr["dtype", IsEqual[Int64]]
DTypeString = IsAttr["dtype", IsEqual[string]]



import datetime as dt
from enum import Enum, auto
from functools import partial
from pathlib import Path
from typing import Any, cast

from hypothesis import given, settings
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    sampled_from,
    times,
)
from luigi import BoolParameter, Task
from luigi.notifications import smtp

from utilities.hypothesis.luigi import namespace_mixins
from utilities.luigi import (
    EnumParameter,
    PathTarget,
    TimeParameter,
    _yield_task_classes,
    build,
    clone,
    get_dependencies_downstream,
    get_dependencies_upstream,
    get_task_classes,
)


class TestBuild:
    @given(namespace_mixin=namespace_mixins())
    def test_main(self, namespace_mixin: Any) -> None:
        class Example(namespace_mixin, Task):
            ...

        _ = build([Example()], local_scheduler=True)


class TestClone:
    @given(namespace_mixin=namespace_mixins(), truth=booleans())
    def test_main(self, namespace_mixin: Any, truth: bool) -> None:
        class A(namespace_mixin, Task):
            truth = cast(bool, BoolParameter())

        class B(namespace_mixin, Task):
            truth = cast(bool, BoolParameter())

        a = A(truth)
        result = clone(a, B)
        expected = B(truth)
        assert result is expected


class TestEnumParameter:
    @given(data=data())
    def test_main(self, data: DataObject) -> None:
        class Example(Enum):
            member = auto()

        param = EnumParameter(Example)
        input_ = data.draw(sampled_from([Example.member, "member"]))
        norm = param.normalize(input_)
        assert param.parse(param.serialize(norm)) == norm


class TestGetDependencies:
    @given(namespace_mixin=namespace_mixins())
    @settings(max_examples=1)
    def test_main(self, namespace_mixin: Any) -> None:
        class A(namespace_mixin, Task):
            ...

        class B(namespace_mixin, Task):
            def requires(self) -> A:
                return clone(self, A)

        class C(namespace_mixin, Task):
            def requires(self) -> B:
                return clone(self, B)

        a, b, c = A(), B(), C()
        ((up_a, down_a), (up_b, down_b), (up_c, down_c)) = map(
            self._get_sets,
            [a, b, c],
        )
        assert up_a == set()
        assert down_a == {b}
        assert up_b == {a}
        assert down_b == {c}
        assert up_c == {b}
        assert down_c == set()

        (
            (up_a_rec, down_a_rec),
            (up_b_rec, down_b_rec),
            (up_c_rec, down_c_rec),
        ) = map(partial(self._get_sets, recursive=True), [a, b, c])
        assert up_a_rec == set()
        assert down_a_rec == {b, c}
        assert up_b_rec == {a}
        assert down_b_rec == {c}
        assert up_c_rec == {a, b}
        assert down_c_rec == set()

    @staticmethod
    def _get_sets(
        task: Task,
        /,
        *,
        recursive: bool = False,
    ) -> tuple[set[Task], set[Task]]:
        return set(get_dependencies_upstream(task, recursive=recursive)), set(
            get_dependencies_downstream(task, recursive=recursive),
        )


class TestGetTaskClasses:
    @given(namespace_mixin=namespace_mixins())
    @settings(max_examples=1)
    def test_main(self, namespace_mixin: Any) -> None:
        class Example(namespace_mixin, Task):
            ...

        assert Example in get_task_classes()

    def test_notifications(self) -> None:
        assert smtp not in _yield_task_classes()

    @given(namespace_mixin=namespace_mixins())
    @settings(max_examples=1)
    def test_filter(self, namespace_mixin: Any) -> None:
        class Parent(namespace_mixin, Task):
            ...

        class Child(Parent):
            ...

        result = get_task_classes(cls=Parent)
        expected = frozenset([Child])
        assert result == expected


class TestPathTarget:
    def test_main(self, tmp_path: Path) -> None:
        target = PathTarget(path := tmp_path.joinpath("file"))
        assert isinstance(target.path, Path)
        assert not target.exists()
        path.touch()
        assert target.exists()


class TestTimeParameter:
    @given(time=times())
    def test_main(self, time: dt.time) -> None:
        param = TimeParameter()
        norm = param.normalize(time)
        assert param.parse(param.serialize(norm)) == time



import datetime as dt
from typing import Any, cast

from hypothesis import assume, given
from hypothesis.extra.pandas import range_indexes
from hypothesis.strategies import integers
from pandas import (
    DataFrame,
    Index,
    NaT,
    RangeIndex,
    Series,
    Timestamp,
    to_datetime,
)
from pytest import mark, param, raises

from utilities.datetime import UTC
from utilities.hypothesis import text_ascii
from utilities.pandas import (
    TIMESTAMP_MAX_AS_DATE,
    TIMESTAMP_MAX_AS_DATETIME,
    TIMESTAMP_MIN_AS_DATE,
    TIMESTAMP_MIN_AS_DATETIME,
    DataFrameRangeIndexError,
    Int64,
    RangeIndexNameError,
    RangeIndexStartError,
    RangeIndexStepError,
    SeriesRangeIndexError,
    boolean,
    check_range_index,
    string,
    timestamp_to_date,
    timestamp_to_datetime,
)


class TestCheckRangeIndex:
    @given(index=range_indexes())
    def test_main(self, index: RangeIndex) -> None:
        check_range_index(index)

    def test_type(self) -> None:
        index = Index([], dtype=float)
        with raises(TypeError):
            check_range_index(index)

    @given(start=integers(-10, 10), stop=integers(-10, 10))
    def test_start(self, start: int, stop: int) -> None:
        _ = assume(start != 0)
        index = RangeIndex(start=start, stop=stop)
        with raises(RangeIndexStartError):
            check_range_index(index)

    @given(step=integers(-10, 10))
    def test_step(self, step: int) -> None:
        _ = assume(step not in {0, 1})
        index = RangeIndex(step=step)
        with raises(RangeIndexStepError):
            check_range_index(index)

    @given(index=range_indexes(name=text_ascii()))
    def test_name(self, index: RangeIndex) -> None:
        with raises(RangeIndexNameError):
            check_range_index(index)

    def test_series_pass(self) -> None:
        series = Series(index=RangeIndex(0), dtype=float)
        check_range_index(series)

    def test_series_fail(self) -> None:
        series = Series(dtype=float)
        with raises(SeriesRangeIndexError):
            check_range_index(series)

    def test_dataframe_pass(self) -> None:
        df = DataFrame(index=RangeIndex(0))
        check_range_index(df)

    def test_dataframe_fail(self) -> None:
        df = DataFrame()
        with raises(DataFrameRangeIndexError):
            check_range_index(df)


class TestDTypes:
    @mark.parametrize("dtype", [param(Int64), param(boolean), param(string)])
    def test_main(self, dtype: Any) -> None:
        assert isinstance(Series([], dtype=dtype), Series)


class TestTimestampMinMaxAsDate:
    def test_min(self) -> None:
        date = TIMESTAMP_MIN_AS_DATE
        assert isinstance(to_datetime(cast(Timestamp, date)), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(cast(Timestamp, date - dt.timedelta(days=1)))

    def test_max(self) -> None:
        date = TIMESTAMP_MAX_AS_DATE
        assert isinstance(to_datetime(cast(Timestamp, date)), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(cast(Timestamp, date + dt.timedelta(days=1)))


class TestTimestampMinMaxAsDateTime:
    def test_min(self) -> None:
        date = TIMESTAMP_MIN_AS_DATETIME
        assert isinstance(to_datetime(date), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(date - dt.timedelta(microseconds=1))

    def test_max(self) -> None:
        date = TIMESTAMP_MAX_AS_DATETIME
        assert isinstance(to_datetime(date), Timestamp)
        with raises(ValueError, match="Out of bounds nanosecond timestamp"):
            _ = to_datetime(date + dt.timedelta(microseconds=1))


class TestTimestampToDate:
    @mark.parametrize(
        ("timestamp", "expected"),
        [
            param(to_datetime("2000-01-01"), dt.date(2000, 1, 1)),
            param(to_datetime("2000-01-01 12:00:00"), dt.date(2000, 1, 1)),
        ],
    )
    def test_main(self, timestamp: Any, expected: dt.date) -> None:
        assert timestamp_to_date(timestamp) == expected

    def test_error(self) -> None:
        with raises(ValueError, match="Invalid value"):
            _ = timestamp_to_date(NaT)


class TestTimestampToDateTime:
    @mark.parametrize(
        ("timestamp", "expected"),
        [
            param(
                to_datetime("2000-01-01"),
                dt.datetime(2000, 1, 1, tzinfo=UTC),
            ),
            param(
                to_datetime("2000-01-01 12:00:00"),
                dt.datetime(2000, 1, 1, 12, tzinfo=UTC),
            ),
            param(
                to_datetime("2000-01-01 12:00:00+00:00"),
                dt.datetime(2000, 1, 1, 12, tzinfo=UTC),
            ),
        ],
    )
    def test_main(self, timestamp: Any, expected: dt.datetime) -> None:
        assert timestamp_to_datetime(timestamp) == expected

    def test_error(self) -> None:
        with raises(ValueError, match="Invalid value"):
            _ = timestamp_to_datetime(NaT)



from collections.abc import Iterable, Iterator
from logging import getLogger
from pathlib import Path

from attrs import asdict
from beartype import beartype
from click import command

from utilities.logging import basic_config
from utilities.re import NoMatchesError, extract_group
from utilities.rotate_logs.classes import Config, Item
from utilities.typed_settings import click_options

_CONFIG = Config()
_LOGGER = getLogger(__name__)


@command()
@click_options(Config)
@beartype
def main(config: Config, /) -> None:
    """CLI for the `rotate_logs` script."""
    basic_config()
    _log_config(config)
    item_lists = _yield_items(
        path=config.path,
        extension=config.extension,
        size=config.size,
    )
    if config.dry_run:
        for items in item_lists:
            for item in items:
                _LOGGER.debug("%s", item)
    else:
        for items in item_lists:
            _rotate_items(items, keep=config.keep)


@beartype
def _log_config(config: Config, /) -> None:
    for key, value in asdict(config).items():
        _LOGGER.info("%-9s = %s", key, value)


@beartype
def _yield_items(
    *,
    path: Path = _CONFIG.path,
    extension: str = _CONFIG.extension,
    size: int = _CONFIG.size,
) -> Iterator[frozenset[Item]]:
    for p in path.rglob("*"):
        if p.suffix == f".{extension}" and p.stat().st_size >= size:
            yield frozenset(_yield_for_head(p, extension=extension))


@beartype
def _yield_for_head(
    path: Path,
    /,
    *,
    extension: str = _CONFIG.extension,
) -> Iterator[Item]:
    yield Item(path, path)
    pattern = rf"^{path.stem}\.{extension}\.(\d+)$"
    for p in path.parent.iterdir():
        try:
            num = extract_group(pattern, p.name)
        except NoMatchesError:
            pass
        else:
            yield Item(p, path, num=int(num))


@beartype
def _rotate_items(
    items: Iterable[Item],
    /,
    *,
    keep: int = _CONFIG.keep,
) -> None:
    for item in sorted(items, key=_key, reverse=True):
        _rotate_item(item, keep=keep)


@beartype
def _key(item: Item, /) -> int:
    return item.num


@beartype
def _rotate_item(item: Item, /, *, keep: int = _CONFIG.keep) -> None:
    path = item.path
    if item.num >= keep:
        _LOGGER.info("Removing file: %s", path)
        path.unlink(missing_ok=True)
    else:
        head = item.head
        new_num = item.num + 1
        new_path = head.with_name(f"{head.name}.{new_num}")
        _LOGGER.info("Rotating file: %s -> %s", path, new_path)
        _ = path.rename(new_path)



from pathlib import Path
from typing import Union

from beartype import beartype

from utilities.re import extract_group

PathLike = Union[Path, str]


@beartype
def ensure_suffix(path: PathLike, suffix: str, /) -> Path:
    """Ensure a path has the required suffix."""
    as_path = Path(path)
    parts = as_path.name.split(".")
    clean_suffix = extract_group(r"^\.(\w+)$", suffix)
    if parts[-1] != clean_suffix:
        parts.append(clean_suffix)
    return as_path.with_name(".".join(parts))



from operator import neg, pow
from typing import Literal

from pytest import mark, param

from utilities.pqdm import pmap, pstarmap


class TestPMap:
    @mark.parametrize("parallelism", [param("processes"), param("threads")])
    @mark.parametrize("n_jobs", [param(1), param(2)])
    def test_unary(
        self,
        parallelism: Literal["processes", "threads"],
        n_jobs: int,
    ) -> None:
        result = pmap(neg, [1, 2, 3], parallelism=parallelism, n_jobs=n_jobs)
        expected = [-1, -2, -3]
        assert result == expected

    @mark.parametrize("parallelism", [param("processes"), param("threads")])
    @mark.parametrize("n_jobs", [param(1), param(2)])
    def test_binary(
        self,
        parallelism: Literal["processes", "threads"],
        n_jobs: int,
    ) -> None:
        result = pmap(
            pow,
            [2, 3, 10],
            [5, 2, 3],
            parallelism=parallelism,
            n_jobs=n_jobs,
        )
        expected = [32, 9, 1000]
        assert result == expected


class TestPStarMap:
    @mark.parametrize("parallelism", [param("processes"), param("threads")])
    @mark.parametrize("n_jobs", [param(1), param(2)])
    def test_unary(
        self,
        parallelism: Literal["processes", "threads"],
        n_jobs: int,
    ) -> None:
        result = pstarmap(
            neg,
            [(1,), (2,), (3,)],
            parallelism=parallelism,
            n_jobs=n_jobs,
        )
        expected = [-1, -2, -3]
        assert result == expected

    @mark.parametrize("parallelism", [param("processes"), param("threads")])
    @mark.parametrize("n_jobs", [param(1), param(2)])
    def test_binary(
        self,
        parallelism: Literal["processes", "threads"],
        n_jobs: int,
    ) -> None:
        result = pstarmap(
            pow,
            [(2, 5), (3, 2), (10, 3)],
            parallelism=parallelism,
            n_jobs=n_jobs,
        )
        expected = [32, 9, 1000]
        assert result == expected



from utilities.class_name import get_class_name


class TestGetClassName:
    def test_class(self) -> None:
        class Example:
            ...

        result = get_class_name(Example)
        expected = "Example"
        assert result == expected

    def test_instance(self) -> None:
        class Example:
            ...

        result = get_class_name(Example())
        expected = "Example"
        assert result == expected



import datetime as dt
from pathlib import Path
from typing import Any

from pytest import mark, param, raises

from utilities.datetime import UTC
from utilities.json import serialize


class TestSerialize:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(dt.date(2000, 1, 1), '"2000-01-01"'),
            param(
                dt.datetime(2000, 1, 1, 12, tzinfo=UTC),
                '"2000-01-01T12:00:00+00:00"',
            ),
            param(Path("a/b/c"), '"a/b/c"'),
            param({1, 2, 3}, '"set([1, 2, 3])"'),
            param({"a", "b", "c"}, '"set([\\"a\\", \\"b\\", \\"c\\"])"'),
            param(frozenset([1, 2, 3]), '"frozenset([1, 2, 3])"'),
            param(
                frozenset(["a", "b", "c"]),
                '"frozenset([\\"a\\", \\"b\\", \\"c\\"])"',
            ),
        ],
    )
    def test_main(self, x: Any, expected: str) -> None:
        assert serialize(x) == expected

    def test_error(self) -> None:
        class Example:
            pass

        with raises(TypeError, match="Invalid type"):
            _ = serialize(Example())



import datetime as dt
from collections.abc import Iterator
from functools import partial
from getpass import getuser
from itertools import islice
from logging import getLogger
from pathlib import Path
from shutil import rmtree
from typing import Optional

from attrs import asdict
from beartype import beartype
from click import command

from utilities.clean_dir.classes import Config, Item
from utilities.datetime import UTC
from utilities.logging import basic_config
from utilities.typed_settings import click_options

_CONFIG = Config()
_LOGGER = getLogger(__name__)


@command()
@click_options(Config)
@beartype
def main(config: Config, /) -> None:
    """CLI for the `clean_dir` script."""
    basic_config()
    _log_config(config)
    if config.dry_run:
        for item in _yield_items(
            path=config.path,
            days=config.days,
            chunk_size=config.chunk_size,
        ):
            _LOGGER.debug("%s", item.path)
    else:
        _clean_dir(
            path=config.path,
            days=config.days,
            chunk_size=config.chunk_size,
        )


@beartype
def _log_config(config: Config, /) -> None:
    for key, value in asdict(config).items():
        _LOGGER.info("%-10s = %s", key, value)


@beartype
def _clean_dir(
    *,
    path: Path = _CONFIG.path,
    days: int = _CONFIG.days,
    chunk_size: Optional[int] = _CONFIG.chunk_size,
) -> None:
    while True:
        iterator = _yield_items(path=path, days=days, chunk_size=chunk_size)
        if len(items := list(iterator)) >= 1:
            for item in items:
                item.clean()
        else:
            return


@beartype
def _yield_items(
    *,
    path: Path = _CONFIG.path,
    days: int = _CONFIG.days,
    chunk_size: Optional[int] = _CONFIG.chunk_size,
) -> Iterator[Item]:
    it = _yield_inner(path=path, days=days)
    if chunk_size is not None:
        return islice(it, chunk_size)
    return it


@beartype
def _yield_inner(
    *,
    path: Path = _CONFIG.path,
    days: int = _CONFIG.days,
) -> Iterator[Item]:
    for p in path.rglob("*"):
        yield from _yield_from_path(p, path=path, days=days)


@beartype
def _yield_from_path(
    p: Path,
    /,
    *,
    path: Path = _CONFIG.path,
    days: int = _CONFIG.days,
) -> Iterator[Item]:
    if p.is_symlink():
        yield from _yield_from_path(p.resolve(), path=path, days=days)
    elif _is_owned_and_relative(p, path=path):  # pragma: no cover
        if (p.is_file() or p.is_socket()) and _is_old(p, days=days):
            yield Item(p, partial(_unlink_path, p))
        elif p.is_dir() and _is_empty(p):
            yield Item(p, partial(_unlink_dir, p))


@beartype
def _is_owned_and_relative(p: Path, /, *, path: Path = _CONFIG.path) -> bool:
    try:
        return (p.owner() == getuser()) and p.is_relative_to(path)
    except FileNotFoundError:  # pragma: no cover
        return False


@beartype
def _is_empty(path: Path, /) -> bool:
    return len(list(path.iterdir())) == 0


@beartype
def _is_old(path: Path, /, *, days: int = _CONFIG.days) -> bool:
    age = dt.datetime.now(tz=UTC) - dt.datetime.fromtimestamp(
        path.stat().st_mtime,
        tz=UTC,
    )
    return age >= dt.timedelta(days=days)


@beartype
def _unlink_path(path: Path, /) -> None:
    _LOGGER.info("Removing file:      %s", path)
    path.unlink(missing_ok=True)


@beartype
def _unlink_dir(path: Path, /) -> None:
    _LOGGER.info("Removing directory: %s", path)
    rmtree(path)



"""Utilities for text objects."""


from textwrap import dedent
from typing import Any

from beartype import beartype


@beartype
def ensure_str(x: Any, /) -> str:
    """Ensure an object is a string."""
    if isinstance(x, str):
        return x
    msg = f"{x=}"
    raise NotAStringError(msg)


class NotAStringError(TypeError):
    """Raised when an object is not a string."""


@beartype
def strip_and_dedent(text: str, /) -> str:
    """Strip and dedent a string."""
    return dedent(text.strip("\n")).strip("\n")



from typing import Any

from attrs import asdict
from beartype import beartype
from xarray import DataArray


@beartype
def rename_data_arrays(obj: Any, /) -> None:
    """Rename the arrays on a field."""
    for key, value in asdict(obj).items():
        if isinstance(value, DataArray) and (value.name != key):
            setattr(obj, key, value.rename(key))



from typing import Any


class TestCheck:
    def test_regular(self, testdir: Any) -> None:
        testdir.makepyfile(
            """
            from utilities.pytest_check import check

            def test_main():
                with check():
                    assert False, "first"
                with check():
                    assert False, "second"
            """,
        )
        result = testdir.runpytest()
        result.assert_outcomes(failed=1)
        result.stdout.fnmatch_lines(
            [
                "FAILURE: first",
                "FAILURE: second",
                "Failed Checks: 2",
            ],
        )

    def test_fail_on_first(self, testdir: Any) -> None:
        testdir.makepyfile(
            """
            from utilities.os import temp_environ
            from utilities.pytest_check import check

            def test_main():
                with temp_environ({"PYTEST_CHECK": "0"}):
                    with check():
                        assert False, "first"
                    with check():
                        assert False, "second"
            """,
        )
        result = testdir.runpytest()
        result.assert_outcomes(failed=1)
        result.stdout.re_match_lines([".*AssertionError: first"])
